library benchpress.src.metric.perflog_metric;

import "package:angular2/src/facade/async.dart"
    show PromiseWrapper, Future, TimerWrapper;
import "package:angular2/src/facade/lang.dart"
    show isPresent, isBlank, StringWrapper, Math, RegExpWrapper, NumberWrapper;
import "package:angular2/src/facade/exceptions.dart"
    show BaseException, WrappedException;
import "package:angular2/src/facade/collection.dart"
    show ListWrapper, StringMapWrapper;
import "package:angular2/src/core/di.dart"
    show bind, provide, Provider, OpaqueToken;
import "../web_driver_extension.dart" show WebDriverExtension, PerfLogFeatures;
import "../metric.dart" show Metric;
import "../common_options.dart" show Options;

/**
 * A metric that reads out the performance log
 */
class PerflogMetric extends Metric {
  WebDriverExtension _driverExtension;
  Function _setTimeout;
  Map<String, dynamic> _microMetrics;
  bool _forceGc;
  bool _captureFrames;
  // TODO(tbosch): use static values when our transpiler supports them
  static List<Provider> get BINDINGS {
    return _PROVIDERS;
  }

  // TODO(tbosch): use static values when our transpiler supports them
  static OpaqueToken get SET_TIMEOUT {
    return _SET_TIMEOUT;
  }

  List<Map<String, dynamic>> _remainingEvents;
  num _measureCount;
  PerfLogFeatures _perfLogFeatures;
  /**
   * @param driverExtension
   * @param setTimeout
   * @param microMetrics Name and description of metrics provided via console.time / console.timeEnd
   **/
  PerflogMetric(this._driverExtension, this._setTimeout, this._microMetrics,
      this._forceGc, this._captureFrames)
      : super() {
    /* super call moved to initializer */;
    this._remainingEvents = [];
    this._measureCount = 0;
    this._perfLogFeatures = _driverExtension.perfLogFeatures();
  }
  Map<String, dynamic> describe() {
    var res = {
      "scriptTime": "script execution time in ms, including gc and render",
      "pureScriptTime": "script execution time in ms, without gc nor render"
    };
    if (this._perfLogFeatures.render) {
      res["renderTime"] = "render time in ms";
    }
    if (this._perfLogFeatures.gc) {
      res["gcTime"] = "gc time in ms";
      res["gcAmount"] = "gc amount in kbytes";
      res["majorGcTime"] = "time of major gcs in ms";
      if (this._forceGc) {
        res["forcedGcTime"] = "forced gc time in ms";
        res["forcedGcAmount"] = "forced gc amount in kbytes";
      }
    }
    if (this._captureFrames) {
      if (!this._perfLogFeatures.frameCapture) {
        var warningMsg =
            "WARNING: Metric requested, but not supported by driver";
        // using dot syntax for metric name to keep them grouped together in console reporter
        res["frameTime.mean"] = warningMsg;
        res["frameTime.worst"] = warningMsg;
        res["frameTime.best"] = warningMsg;
        res["frameTime.smooth"] = warningMsg;
      } else {
        res["frameTime.mean"] =
            "mean frame time in ms (target: 16.6ms for 60fps)";
        res["frameTime.worst"] = "worst frame time in ms";
        res["frameTime.best"] = "best frame time in ms";
        res["frameTime.smooth"] = "percentage of frames that hit 60fps";
      }
    }
    StringMapWrapper.forEach(this._microMetrics, (desc, name) {
      StringMapWrapper.set(res, name, desc);
    });
    return res;
  }

  Future<dynamic> beginMeasure() {
    var resultPromise = PromiseWrapper.resolve(null);
    if (this._forceGc) {
      resultPromise = resultPromise.then((_) => this._driverExtension.gc());
    }
    return resultPromise.then((_) => this._beginMeasure());
  }

  Future<Map<String, dynamic>> endMeasure(bool restart) {
    if (this._forceGc) {
      return this._endPlainMeasureAndMeasureForceGc(restart);
    } else {
      return this._endMeasure(restart);
    }
  }

  _endPlainMeasureAndMeasureForceGc(bool restartMeasure) {
    return this._endMeasure(true).then((measureValues) {
      // disable frame capture for measurments during forced gc
      var originalFrameCaptureValue = this._captureFrames;
      this._captureFrames = false;
      return this
          ._driverExtension
          .gc()
          .then((_) => this._endMeasure(restartMeasure))
          .then((forceGcMeasureValues) {
        this._captureFrames = originalFrameCaptureValue;
        StringMapWrapper.set(
            measureValues, "forcedGcTime", forceGcMeasureValues["gcTime"]);
        StringMapWrapper.set(
            measureValues, "forcedGcAmount", forceGcMeasureValues["gcAmount"]);
        return measureValues;
      });
    });
  }

  Future<dynamic> _beginMeasure() {
    return this
        ._driverExtension
        .timeBegin(this._markName(this._measureCount++));
  }

  Future<Map<String, dynamic>> _endMeasure(bool restart) {
    var markName = this._markName(this._measureCount - 1);
    var nextMarkName = restart ? this._markName(this._measureCount++) : null;
    return this
        ._driverExtension
        .timeEnd(markName, nextMarkName)
        .then((_) => this._readUntilEndMark(markName));
  }

  _readUntilEndMark(String markName, [num loopCount = 0, startEvent = null]) {
    if (loopCount > _MAX_RETRY_COUNT) {
      throw new BaseException(
          '''Tried too often to get the ending mark: ${ loopCount}''');
    }
    return this._driverExtension.readPerfLog().then((events) {
      this._addEvents(events);
      var result = this._aggregateEvents(this._remainingEvents, markName);
      if (isPresent(result)) {
        this._remainingEvents = events;
        return result;
      }
      var completer = PromiseWrapper.completer();
      this._setTimeout(
          () => completer
              .resolve(this._readUntilEndMark(markName, loopCount + 1)),
          100);
      return completer.promise;
    });
  }

  _addEvents(List<Map<String, String>> events) {
    var needSort = false;
    events.forEach((event) {
      if (StringWrapper.equals(event["ph"], "X")) {
        needSort = true;
        var startEvent = {};
        var endEvent = {};
        StringMapWrapper.forEach(event, (value, prop) {
          startEvent[prop] = value;
          endEvent[prop] = value;
        });
        startEvent["ph"] = "B";
        endEvent["ph"] = "E";
        endEvent["ts"] = startEvent["ts"] + startEvent["dur"];
        this._remainingEvents.add(startEvent);
        this._remainingEvents.add(endEvent);
      } else {
        this._remainingEvents.add(event);
      }
    });
    if (needSort) {
      // Need to sort because of the ph==='X' events
      ListWrapper.sort(this._remainingEvents, (a, b) {
        var diff = a["ts"] - b["ts"];
        return diff > 0 ? 1 : diff < 0 ? -1 : 0;
      });
    }
  }

  Map<String, dynamic> _aggregateEvents(
      List<Map<String, dynamic>> events, markName) {
    var result = {"scriptTime": 0, "pureScriptTime": 0};
    if (this._perfLogFeatures.gc) {
      result["gcTime"] = 0;
      result["majorGcTime"] = 0;
      result["gcAmount"] = 0;
    }
    if (this._perfLogFeatures.render) {
      result["renderTime"] = 0;
    }
    if (this._captureFrames) {
      result["frameTime.mean"] = 0;
      result["frameTime.best"] = 0;
      result["frameTime.worst"] = 0;
      result["frameTime.smooth"] = 0;
    }
    StringMapWrapper.forEach(this._microMetrics, (desc, name) {
      result[name] = 0;
    });
    var markStartEvent = null;
    var markEndEvent = null;
    var gcTimeInScript = 0;
    var renderTimeInScript = 0;
    var frameTimestamps = [];
    var frameTimes = [];
    var frameCaptureStartEvent = null;
    var frameCaptureEndEvent = null;
    Map<String, dynamic> intervalStarts = {};
    Map<String, num> intervalStartCount = {};
    events.forEach((event) {
      var ph = event["ph"];
      var name = event["name"];
      var microIterations = 1;
      var microIterationsMatch =
          RegExpWrapper.firstMatch(_MICRO_ITERATIONS_REGEX, name);
      if (isPresent(microIterationsMatch)) {
        name = microIterationsMatch[1];
        microIterations = NumberWrapper.parseInt(microIterationsMatch[2], 10);
      }
      if (StringWrapper.equals(ph, "b") &&
          StringWrapper.equals(name, markName)) {
        markStartEvent = event;
      } else if (StringWrapper.equals(ph, "e") &&
          StringWrapper.equals(name, markName)) {
        markEndEvent = event;
      }
      if (isPresent(markStartEvent) &&
          isBlank(markEndEvent) &&
          identical(event["pid"], markStartEvent["pid"])) {
        if (StringWrapper.equals(ph, "b") &&
            StringWrapper.equals(name, _MARK_NAME_FRAME_CAPUTRE)) {
          if (isPresent(frameCaptureStartEvent)) {
            throw new BaseException(
                "can capture frames only once per benchmark run");
          }
          if (!this._captureFrames) {
            throw new BaseException(
                "found start event for frame capture, but frame capture was not requested in benchpress");
          }
          frameCaptureStartEvent = event;
        } else if (StringWrapper.equals(ph, "e") &&
            StringWrapper.equals(name, _MARK_NAME_FRAME_CAPUTRE)) {
          if (isBlank(frameCaptureStartEvent)) {
            throw new BaseException("missing start event for frame capture");
          }
          frameCaptureEndEvent = event;
        }
        if (StringWrapper.equals(ph, "I") || StringWrapper.equals(ph, "i")) {
          if (isPresent(frameCaptureStartEvent) &&
              isBlank(frameCaptureEndEvent) &&
              StringWrapper.equals(name, "frame")) {
            frameTimestamps.add(event["ts"]);
            if (frameTimestamps.length >= 2) {
              frameTimes.add(frameTimestamps[frameTimestamps.length - 1] -
                  frameTimestamps[frameTimestamps.length - 2]);
            }
          }
        }
        if (StringWrapper.equals(ph, "B") || StringWrapper.equals(ph, "b")) {
          if (isBlank(intervalStarts[name])) {
            intervalStartCount[name] = 1;
            intervalStarts[name] = event;
          } else {
            intervalStartCount[name]++;
          }
        } else if ((StringWrapper.equals(ph, "E") ||
                StringWrapper.equals(ph, "e")) &&
            isPresent(intervalStarts[name])) {
          intervalStartCount[name]--;
          if (identical(intervalStartCount[name], 0)) {
            var startEvent = intervalStarts[name];
            var duration = (event["ts"] - startEvent["ts"]);
            intervalStarts[name] = null;
            if (StringWrapper.equals(name, "gc")) {
              result["gcTime"] += duration;
              var amount = (startEvent["args"]["usedHeapSize"] -
                      event["args"]["usedHeapSize"]) /
                  1000;
              result["gcAmount"] += amount;
              var majorGc = event["args"]["majorGc"];
              if (isPresent(majorGc) && majorGc) {
                result["majorGcTime"] += duration;
              }
              if (isPresent(intervalStarts["script"])) {
                gcTimeInScript += duration;
              }
            } else if (StringWrapper.equals(name, "render")) {
              result["renderTime"] += duration;
              if (isPresent(intervalStarts["script"])) {
                renderTimeInScript += duration;
              }
            } else if (StringWrapper.equals(name, "script")) {
              result["scriptTime"] += duration;
            } else if (isPresent(this._microMetrics[name])) {
              result[name] += duration / microIterations;
            }
          }
        }
      }
    });
    if (!isPresent(markStartEvent) || !isPresent(markEndEvent)) {
      // not all events have been received, no further processing for now
      return null;
    }
    if (isPresent(markEndEvent) &&
        isPresent(frameCaptureStartEvent) &&
        isBlank(frameCaptureEndEvent)) {
      throw new BaseException("missing end event for frame capture");
    }
    if (this._captureFrames && isBlank(frameCaptureStartEvent)) {
      throw new BaseException(
          "frame capture requested in benchpress, but no start event was found");
    }
    if (frameTimes.length > 0) {
      this._addFrameMetrics(result, frameTimes);
    }
    result["pureScriptTime"] =
        result["scriptTime"] - gcTimeInScript - renderTimeInScript;
    return result;
  }

  _addFrameMetrics(Map<String, dynamic> result, List<dynamic> frameTimes) {
    result["frameTime.mean"] =
        frameTimes.fold(0, (a, b) => a + b) / frameTimes.length;
    var firstFrame = frameTimes[0];
    result["frameTime.worst"] =
        frameTimes.fold(firstFrame, (a, b) => a > b ? a : b);
    result["frameTime.best"] =
        frameTimes.fold(firstFrame, (a, b) => a < b ? a : b);
    result["frameTime.smooth"] = frameTimes
            .where((t) => t < _FRAME_TIME_SMOOTH_THRESHOLD)
            .toList()
            .length /
        frameTimes.length;
  }

  _markName(index) {
    return '''${ _MARK_NAME_PREFIX}${ index}''';
  }
}

var _MICRO_ITERATIONS_REGEX = new RegExp(r'(.+)\*(\d+)$');
var _MAX_RETRY_COUNT = 20;
var _MARK_NAME_PREFIX = "benchpress";
var _SET_TIMEOUT = new OpaqueToken("PerflogMetric.setTimeout");
var _MARK_NAME_FRAME_CAPUTRE = "frameCapture";
// using 17ms as a somewhat looser threshold, instead of 16.6666ms
var _FRAME_TIME_SMOOTH_THRESHOLD = 17;
var _PROVIDERS = [
  bind(PerflogMetric).toFactory(
      (driverExtension, setTimeout, microMetrics, forceGc, captureFrames) =>
          new PerflogMetric(driverExtension, setTimeout, microMetrics, forceGc,
              captureFrames),
      [
    WebDriverExtension,
    _SET_TIMEOUT,
    Options.MICRO_METRICS,
    Options.FORCE_GC,
    Options.CAPTURE_FRAMES
  ]),
  provide(_SET_TIMEOUT,
      useValue: (fn, millis) => TimerWrapper.setTimeout(fn, millis))
];
